/*
 * Factor integers into their prime factors.
 * Works on up to 14 digit numbers.
 *
 * Written by George Gesslein II.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <time.h>
#include <assert.h>

double nn, vv;
double huge_value = 1.0e14;
double sq[] = {
	10, 2, 4, 2, 4, 6, 2, 6,
	 4, 2, 4, 6, 6, 2, 6, 4,
	 2, 6, 4, 6, 8, 4, 2, 4,
	 2, 4, 8, 6, 4, 6, 2, 4,
	 6, 2, 6, 6, 4, 2, 4, 6,
	 2, 6, 4, 2, 4, 2,10, 2
};

double	unique[64];
int	ucnt[64];
int	utmp[64];
int	uno;

main(argc, argv)
int	argc;
char	*argv[];
{
	int		i, j;
	int		ind;
	double		junk, temp;
	double		ii;
	double		start;
	double		divisor;
	double		d1, d2;
	long		count;

	ind = 1;
	for (;; ind++) {
		if (argc > 1) {
			if (ind >= argc)
				break;
			if (sscanf(argv[ind], "%lf", &nn) != 1) {
				printf("Arguments to this program must be numbers!\n");
				exit(1);
			}
		} else {
			if (isatty(0)) {
				printf("Enter integer to factor (0 to exit): ");
			}
			if (scanf("%lf", &nn) != 1)
				break;
		}
		start = nn;
		if (nn == 0.0)
			break;
		if (fabs(nn) > huge_value) {
			printf("Number is too large!\n");
			continue;
		}
		if (modf(nn, &junk) != 0.0) {
			printf("%lg is not an integer!\n", nn);
			continue;
		}
		printf("The prime factors of %.0f are:", nn);
		uno = 0;
		vv = 1.0 + sqrt(fabs(nn));
		try_factor(2.0);
		try_factor(3.0);
		try_factor(5.0);
		try_factor(7.0);
		ii = 1.0;
		while (ii <= vv) {
			for (j = 0; j < 48; j++) {
				ii += sq[j];
				try_factor(ii);
			}
		}
		assert(modf(nn, &junk) == 0.0);
		if (nn != 1.0) {
			try_factor(nn);
		}
		if (start == 1.0) {
			printf(" none");
		}
		printf(";\n");
		if (uno > 0 && start > 0.0) {
			for (i = 0; i < uno; i++) {
				utmp[i] = 0;
			}
			temp = 0.0;
			divisor = 1.0;
			count = 0;
			do {
				count++;
				temp += divisor;
				divisor = 1.0;
				inc_utmp();
				for (i = 0; i < uno; i++) {
					for (j = 0; j < utmp[i]; j++) {
						divisor *= unique[i];
					}
				}
			} while (divisor < start);
			assert(divisor == start);
			printf("The number of divisors = %ld;  ", count);
			printf("The sum of the divisors = %.0f;\n", temp);
			d1 = start;
			d2 = 0;
			do {
				d2 += fmod(d1, 10.0);
				d1 /= 10;
				modf(d1, &d1);
			} while (d1 >= 1.0);
			printf("The sum of the digits = %.0f;\n", d2);
			printf("The number in hexadecimal: ");
			printbase(start, 16.0);
			printf(";\nThe number in binary: ");
			printbase(start, 2.0);
			printf(";\n");
			if (temp == start) {
				printf("PERFECT NUMBER!\n");
			} else if (fabs(temp - start) == 1.0) {
				printf("Missed being a perfect number by 1!\n");
			}
			if (temp <= 1.0)
				printf("Prime number!\n");
		}
		printf("\n");
	}
	exit(0);
}

/*
 * Display positive number "num" in base "base".
 */
printbase(num, base)
double	num, base;
{
	int	i;
	char	ch;

	num = floor(num);
	if (num == 0.0) {
		return;
	}
	i = fmod(num, base);
	printbase(num / base, base);
	if (i < 10) {
		ch = '0' + i;
	} else {
		ch = 'A' + (i - 10);
	}
	printf("%c", ch);
}

/*
 * See if "arg" is one or more factors of "nn".
 * If so, save it, print it, and remove it from "nn".
 */
try_factor(arg)
double	arg;
{
	for (;;) {
		if (fmod(nn, arg) == 0.0) {
			if (uno > 0 && unique[uno-1] == arg) {
				ucnt[uno-1]++;
			} else {
				unique[uno] = arg;
				ucnt[uno] = 1;
				uno++;
			}
			printf(" %.0f", arg);
			nn = nn/arg;
			vv = 1.0 + sqrt(fabs(nn));
			if (nn <= 1.0 && nn >= -1.0)
				break;
		} else {
			break;
		}
	}
}

inc_utmp()
{
	int	i;

	for (i = 0;; i++) {
		assert(i < uno);
		utmp[i]++;
		if (utmp[i] <= ucnt[i]) {
			break;
		}
		utmp[i] = 0;
	}
}
